#ifndef MCL_QUICKSORT
#define MCL_QUICKSORT

#include <stdexcept>
#include <cassert>
#include <functional>
#include <algorithm>
#include <list>

#include "MCL/mutex.h"

#include "MCL/algorithms/sequentialSortTask.h"
#include "MCL/data_structures/subview.h"
#include "MCL/data_structures/parallelScanView.h"
#include "MCL/data_structures/blockedView.h"

namespace mcl {
  /**
   * Neutralize two blocks
   * if the low block is NOT neutralized -1 is returned
   * if the high block is NOT neutralized 1 is returned
   * if both blocks are neutralized 0 ir returned
   **/
  template<class BaseView, class Compare>
  int
  Neutralize(SubView<BaseView>& low, 
	     SubView<BaseView>& high, 
	     typename BaseView::value_type& pivot,
	     Compare& cmp){
    typename SubView<BaseView>::iterator lowIter = low.begin();
    typename SubView<BaseView>::iterator highIter = high.begin();
    while(lowIter!=low.end()){
      if(!cmp(*lowIter,pivot))
	break;
      ++lowIter;
    }
    while(highIter!=high.end()){
      if(cmp(*highIter,pivot))
	break;
      ++highIter;
    }
    while(lowIter!=low.end() && highIter!=high.end()){
      std::swap(*lowIter,*highIter);
      while(cmp(*lowIter,pivot)){
	++lowIter;
	if(lowIter == low.end())
	  break;
      }
      while(!cmp(*highIter,pivot)){
	++highIter;
	if(highIter == high.end())
	  break;
      }
    }
    if(lowIter != low.end())
      return -1;
    if(highIter != high.end())
      return 1;
    return 0;
  }

  static Mutex printout_;

  template<class View, class Compare, class Task>
  class ParallelPivotPartitionPhaseTask : public Task {
    typedef ParallelScanView<BlockedView<SubView<View> > > ScanView;
    typedef typename BasicArray<typename ScanView::iterator>::iterator UnNuteredIter;
  public:
    ParallelPivotPartitionPhaseTask(typename View::value_type* p,
				    typename ScanView::iterator start,
				    typename ScanView::iterator end,
				    UnNuteredIter u,
				    Compare cmp)
      : pivot_(p), lowIter_(start), highIter_(end), unNeutralizedIter_(u), highNeutralized_(0), lowNeutralized_(0), cmp_(cmp){}

    void SetUnneutralized(typename ScanView::iterator unneutralized){
      *unNeutralizedIter_ = unneutralized;
    }

    void operator()(){
      bool haveUnneutredlow = true;
      //try to get the first two block to neutralize,
      // if this fails return
      // note that the second getblock cannot succeed if the first failed
      try {	
	lowIter_.GetNextBefore(highIter_);
      } catch(std::range_error& ex) {haveUnneutredlow = false;}
      try {
	highIter_.GetPrevAfter(lowIter_);
      } catch(std::range_error& ex) {
	if (haveUnneutredlow)
	  SetUnneutralized(lowIter_);
	return;
      }
      while(true) {
	SubView<View> highSubView = *highIter_;
	SubView<View> lowSubView = *lowIter_;
	int unnutralized = Neutralize(lowSubView, highSubView, *pivot_, cmp_);
	if (unnutralized < 0) {
	  //lowIter now points to a unneutralized block
	  //highIter is neutralized, so get a new block.
	  ++highNeutralized_;
	  try {
	    highIter_.GetPrevAfter(lowIter_);
	  } catch (std::range_error& ex) {
	    SetUnneutralized(lowIter_);
	    return;
	  }
	} else if (unnutralized > 0) {
	  //highIter is still unneutralize
	  //get a new low block
	  ++lowNeutralized_;
	  try {
	    lowIter_.GetNextBefore(highIter_);
	  } catch (std::range_error& ex) {
	    SetUnneutralized(highIter_);
	    return;
	  }
	} else {
	  //both are neutralized, get two new blocks
	  ++lowNeutralized_;
	  ++highNeutralized_;
	  try {	
	    lowIter_.GetNextBefore(highIter_);
	    haveUnneutredlow = true;
	  } catch(std::range_error& ex) {haveUnneutredlow = false;}
	  try {
	    highIter_.GetPrevAfter(lowIter_);
	  } catch(std::range_error& ex) {
	    if (haveUnneutredlow)
	      SetUnneutralized(lowIter_);
	    return;
	  }
	}
      }
    }
    inline unsigned int GetNumHigh(){
      return highNeutralized_;
    }
    inline unsigned int GetNumLow(){
      return lowNeutralized_;
    }
  protected:
    typename View::value_type* pivot_;
    typename ScanView::iterator lowIter_;
    typename ScanView::iterator highIter_;
    UnNuteredIter unNeutralizedIter_;
    unsigned int highNeutralized_;
    unsigned int lowNeutralized_;
    Compare cmp_;
  };

  template<class View, class Compare, class Task>
  class  ParallelPivotPartitionTask : public Task {
    typedef ParallelScanView<BlockedView<SubView<View> > > ScanView;
  public: 
    ParallelPivotPartitionTask(View* inputView, size_t blockSize, size_t numTasks, typename View::value_type pivot, Compare cmp) 
      : inputView_(inputView,0,inputView->size()), 
	blockSize_(blockSize),
	numTasks_(numTasks),
	pivotPos_(),
	pivot_(pivot),
	cmp_(cmp) {}

    ParallelPivotPartitionTask(SubView<View>& inputView, size_t blockSize, size_t numTasks, typename View::value_type pivot, Compare cmp) 
      : inputView_(inputView), 
	blockSize_(blockSize),
	numTasks_(numTasks),
	pivotPos_(),
	pivot_(pivot),
	cmp_(cmp) {}

    void operator()(){
      BlockedView<SubView<View> > blockedView(&inputView_,blockSize_);
      ScanView scanView(&blockedView);
      size_t numTasksSpawned;
      if (blockedView.size() > 2*numTasks_) {
	numTasksSpawned = numTasks_;
      } else {
	numTasksSpawned = blockedView.size()/2;
      }
      BasicArray<typename ScanView::iterator> potentialUnneutralizedBlocks(numTasksSpawned);
      typename BasicArray<typename ScanView::iterator>::iterator retIter = potentialUnneutralizedBlocks.begin();
      BasicArray<ParallelPivotPartitionPhaseTask<View,Compare,Task>*> tasks(numTasksSpawned);
      for(unsigned int i=0;i<numTasksSpawned;++i){
	tasks[i] = new ParallelPivotPartitionPhaseTask<View,Compare,Task>(&pivot_,
									    scanView.begin(),
									    scanView.end()-1,
									    retIter,
									    cmp_);
	++retIter;
      }
      Spawn(tasks);
      size_t numHighNeutralized = 0;
      size_t numLowNeutralized = 0;
      for(size_t i=0;i<tasks.size();++i){
	numHighNeutralized += tasks[i]->GetNumHigh();
	numLowNeutralized += tasks[i]->GetNumLow();
	delete tasks[i];
      }
      //sequential partition phase
      std::list<typename ScanView::iterator> unneutralizedBlocks;
      typename BasicArray<typename ScanView::iterator>::iterator unneutralizedArrayIter = potentialUnneutralizedBlocks.begin();
      while(unneutralizedArrayIter != potentialUnneutralizedBlocks.end()){
	if((**unneutralizedArrayIter).size() > 0){
	  unneutralizedBlocks.push_back(*unneutralizedArrayIter);
	}
	++unneutralizedArrayIter;
      }
      unneutralizedBlocks.sort();
      //neutralize remaining blocks
      //find the partition boundaries
      //typename std::list<typename ScanView::iterator>::iterator unneutralizedIter = unneutralizedBlocks.begin();
      //lowBoundary is the last low neutralized block
      //high boundary is the first high neutralized block
      typename ScanView::iterator lowBoundary = scanView.begin() + numLowNeutralized-1;
      typename ScanView::iterator highBoundary = scanView.begin() + (scanView.size() - numHighNeutralized);
      typename std::list<typename ScanView::iterator>::iterator low = unneutralizedBlocks.begin();
      typename std::list<typename ScanView::iterator>::iterator high = --unneutralizedBlocks.end();
      {
	//Mutex::Lock l(&printout_);
	//	std::cout<<"high: "<<**high<<" low: "<<**low<<std::endl;
	//std::cout<<"scanview = "<<scanView<<" size:"<<scanView.size()<<std::endl;
      }
      //if any blocks were high neutralized and the last block is unneutralized 
      //neutralize the last block by swapping smo of the elements form a high neutralized block
      //the used high neutralized block is now unneutralized and is added to unneutralized set
      if(*high == scanView.end()-1){
	if(highBoundary != scanView.end()) {
	  typename ScanView::iterator highNeutralizedSwap = scanView.end()-2;
	  for(size_t i=0;i<numTasksSpawned;++i){ //some task must have high neutralized its first high block
	    //insure that the block is neutralized
	    typename std::list<typename ScanView::iterator>::reverse_iterator unneutralizedCompare = unneutralizedBlocks.rbegin();
	    while(unneutralizedCompare != unneutralizedBlocks.rend() && *unneutralizedCompare >= highNeutralizedSwap){
	      if(*unneutralizedCompare == highNeutralizedSwap){
		break;
	      }
	      ++unneutralizedCompare;
	    }
	    //if highNeutralizedSwap is neutralized test if it is high neutralized
	    if(unneutralizedCompare == unneutralizedBlocks.rend() || *unneutralizedCompare < highNeutralizedSwap){
	      if(!cmp_((*highNeutralizedSwap)[0],pivot_)){
		for(size_t j=0;j<(**high).size();++j){
		  std::swap((**high)[j],(*highNeutralizedSwap)[j]);
		}
		unneutralizedBlocks.erase(high);
		unneutralizedBlocks.insert(unneutralizedBlocks.end(),highNeutralizedSwap);
		unneutralizedBlocks.sort();
		break;
	      }
	    }
	  }
	  ++highNeutralizedSwap;
	}
	high = --unneutralizedBlocks.end();
	low = unneutralizedBlocks.begin();
      }
      //neutralize the remaining block
      while(low != high){
	SubView<View> highBlock(**high);
	SubView<View> lowBlock(**low);
	int side = Neutralize(lowBlock,highBlock,pivot_,cmp_);
	if(side != -1){
	  //low block was neutralized
	  if(*low <= lowBoundary){
	    ++lowBoundary;
	    ++numLowNeutralized;
	    typename std::list<typename ScanView::iterator>::iterator del(low);
	    ++low;
	    unneutralizedBlocks.erase(del);
	  } else {
	    ++low;
	  }
	}
	if(side != 1){
	  //high was neutralized
	  bool done(high == low);
	  if(*high >= highBoundary){
	    --highBoundary;
	    ++numHighNeutralized;
	    typename std::list<typename ScanView::iterator>::iterator del(high);
	    --high;
	    unneutralizedBlocks.erase(del);
	    if(done) {
	      //this happens if the the two last blocks are neutralized simultaniously
	      break;
	    } 
	  } else {
	    if(done) {
	      //this happens if the the two last blocks are neutralized simultaniously
	      break;
	    }
	    --high;
	  }
	}
      }
      //at most one block is unneutralized
      typename std::list<typename ScanView::iterator>::iterator remaining = unneutralizedBlocks.end();
      if(low == high) {
	//if low and high are equal a block is still unneutralized
	//if no block is remaining then low > high
	remaining = low;
      }
      //move any wrongly placed blocks to the position between the boundaries
      if(highBoundary != scanView.end()) {
	typename std::list<typename ScanView::iterator>::iterator unneutralizedIter = unneutralizedBlocks.begin();
	while(unneutralizedIter != unneutralizedBlocks.end()) {
	  if(*unneutralizedIter <= lowBoundary){
	    //find a low neutralized block between the boundaries to swap with
	    //because the block is neutralized in sorted order are we sure that the block cannot be low neutralized
	    typename ScanView::iterator swapIter = highBoundary-1;
	    while(swapIter != lowBoundary){
	      if(cmp_((*swapIter)[0],pivot_) && (remaining==unneutralizedBlocks.end() || swapIter!=*remaining)){
		//found a low neutralized block
		(*swapIter).swapElements(**unneutralizedIter);
		break;
	      } 
	      --swapIter;
	    }
	  }
	  if(*unneutralizedIter >= highBoundary){
	    //find a high neutralized block between the boundaries to swap with
	    typename ScanView::iterator swapIter = lowBoundary+1;
	    while(swapIter != highBoundary){
	      if(!cmp_((*swapIter)[0],pivot_) && swapIter != *remaining){
		//found a high neutralized block
		(*swapIter).swapElements(**unneutralizedIter);
		break;
	      } 
	      ++swapIter;
	    }
	  }
	  ++unneutralizedIter;
	}
      }
      //fist element after the low boundary
      typename SubView<View>::iterator lowBoundaryElement(inputView_.begin()+numLowNeutralized*blockSize_);
      //first element in the highBoundary
      typename SubView<View>::iterator highBoundaryElement;
      if(numHighNeutralized == 0){
	highBoundaryElement = inputView_.end();
      } else {
	highBoundaryElement = inputView_.end()-((numHighNeutralized-1)*blockSize_+(*(scanView.end()-1)).size());   
      }
      //partition elements between the boundaries
      std::function<bool(typename View::value_type)> pred 
	= std::bind(cmp_, std::placeholders::_1, pivot_);
      pivotPos_ = std::partition(lowBoundaryElement, highBoundaryElement, pred);
    }

    inline typename View::size_type GetPivotPos(){
      return (pivotPos_ - inputView_.begin());
    }

  protected:
    //view to be partitioned
    SubView<View> inputView_;
    //size of the blocks
    size_t blockSize_;
    //number of block per parallel phase partitioning task
    size_t numTasks_;
    //iterator pointing to the first element in the high part
    typename View::iterator pivotPos_;
    //pivot value
    typename View::value_type pivot_;
    //comparison function
    Compare cmp_;
  };

  template<class T>
  inline const T min(const T f, const T s, const T t) {
    return std::min(std::min(f,s),std::min(s,t));
  }

  template<class T>
  const T max(const T f, const T s, const T t) {
    return std::max(std::max(f,s),std::max(s,t));
  }

  template<class T>
  inline const T median(const T f, const T s, const T t) {
    return (min(f,s,t)+max(f,s,t))/2;
  }

  template<class View, class Compare, class Task>
  class QuickSortTask : public Task {
  public:
    QuickSortTask(SubView<View>& input, size_t grainSize,  size_t blockSize, size_t numTasks, Compare cmp)
      : input_(input), 
	grainSize_(grainSize),
	blockSize_(blockSize),
	numTasks_(numTasks),
	cmp_(cmp) {}
    
    QuickSortTask(View* input, size_t grainSize,  size_t blockSize, size_t numTasks, Compare cmp)
      : input_(input,0,input->size()),
	grainSize_(grainSize),
	blockSize_(blockSize),
	numTasks_(numTasks),
	cmp_(cmp) {}

    void operator()(){
      typename View::value_type pivot = median(input_[0],input_[input_.size()/2],input_[input_.size()-1]);
      //parallel prtitioning of the input
      typename View::size_type pivotPos;
      ParallelPivotPartitionTask<View,Compare,Task> partitionTask(input_, blockSize_, numTasks_, pivot, cmp_);
      Spawn(static_cast<Task*>(&partitionTask));
      pivotPos = partitionTask.GetPivotPos();
    
      SubView<View> low(input_,0,pivotPos);
      SubView<View> high(input_,pivotPos,input_.size());
      Task* lowTask;
      Task* highTask;
      if(low.size() <= grainSize_){
	lowTask = new SortTask<SubView<View>,Compare,Task>(&low,cmp_);
      } else {
	lowTask = new QuickSortTask<View,Compare,Task>(low,grainSize_,blockSize_,numTasks_,cmp_);
      }
      if(high.size() <= grainSize_) {
	highTask = new SortTask<SubView<View>,Compare,Task>(&high,cmp_);
      } else {
	highTask = new QuickSortTask<View,Compare,Task>(high,grainSize_,blockSize_,numTasks_,cmp_);
      }
      Spawn(lowTask,highTask);
      delete lowTask;
      delete highTask;
    }

  private:
    //data to sort
    SubView<View> input_;
    //size of partition that is sorted directly
    size_t grainSize_;
    //number of elements in each block in the partition
    size_t blockSize_;
    //number of blocks per partitiontask
    size_t numTasks_;   
    //comparison function
    Compare cmp_;
  };

}//end namspace mcl
#endif //header guard
