#ifndef MCL_MERGE
#define MCL_MERGE

#include <cmath>
#include <cassert>
#include <map>

//#include "MCL/forEach.h"
#include "MCL/data_structures/subview.h"
#include "MCL/data_structures/blockedView.h"

namespace mcl {

  template<template<class> class TopView, class BaseView, class Compare>
  BasicArray<long int>
  FWayPartition(size_t partitionSize, const TopView<BaseView>& inputView, Compare cmp) {
    mcl::BasicArray<long int> boundary(inputView.size());
    typename BaseView::value_type lMax = inputView[0][0];
    size_t setSize = 0; 
    for(unsigned int i=0;i<inputView.size();++i) {
      setSize += inputView[i].size();
      boundary[i] = -1;
    }
    double partitionFraction = static_cast<double>(partitionSize) / setSize;
    size_t beta = static_cast<size_t>(floor(log2((inputView[0].size() * partitionFraction))));
    size_t step = static_cast<size_t>(pow(2,beta));
    size_t currentSizeGoal
      = static_cast<size_t>(floor(inputView.size() * partitionFraction));
    std::map<typename BaseView::value_type, int, Compare> boundaryHeap(cmp);
    // as views are zero indexed.
    for(size_t i=0;i<inputView.size();++i){
      if(step-1 > inputView[i].size()) {
	continue;
      }
      boundaryHeap[inputView[i][step-1]] = i;
    }
    typename std::map<typename BaseView::value_type, int, Compare>::iterator it
      = boundaryHeap.begin();
    for(size_t i=0;i<currentSizeGoal;++i){
      boundary[(*it).second] += step;
      lMax = (*it).first;
      ++it;
    }
    //inductive steps
    //increase sample
    step = step /2;
    currentSizeGoal = static_cast<size_t>(floor(partitionFraction * floor(inputView[0].size() / static_cast<double>(step))*inputView.size()));
    // step and not step-1 is used as the step is relative to the boundary element.
    while(step >= 1) {
      //reestablish boundary
      for(unsigned int i=0;i<inputView.size();++i){
	if(boundary[i]+step > inputView[i].size())
	  continue;
	if(inputView[i][boundary[i]+step] <= lMax){
	  boundary[i] += step;
	}
      }
      //reestablish size
      //find current partition size
      size_t currentPartitionSize = 0;
      for(unsigned int i=0;i<inputView.size();++i){
	currentPartitionSize += (boundary[i]+1) / step;
      }
      if(currentPartitionSize < currentSizeGoal){
	//increase partition size by moving the boundry indices
	//effectivly add the smallest elements, after the boundary, to the partition.
	boundaryHeap.clear();
	for(unsigned int i=0;i<inputView.size();++i){
	  if(boundary[i]+step<inputView[i].size()){
	    boundaryHeap[inputView[i][boundary[i]+step]] = i;
	  }
	}
	it = boundaryHeap.begin();
	unsigned int numToIncreaseWith = currentSizeGoal - currentPartitionSize;
	for(unsigned int i=0;i<numToIncreaseWith;++i){
	  lMax = (*it).first;
	  unsigned int minIndex = (*it).second;
	  boundaryHeap.erase(it);
	  boundary[minIndex] += step;
	  //if we can move the boundary a step more add the next to the heap
	  if( boundary[minIndex]+step < inputView[minIndex].size()) {
	    boundaryHeap[inputView[minIndex][boundary[minIndex]+step]] = minIndex;
	  }
	  it = boundaryHeap.begin();
	}
      } else if (currentPartitionSize > currentSizeGoal){
	//decrease partition 
	//by moveing the boundary back
	boundaryHeap.clear(); 
	for(unsigned int i=0;i<inputView.size();++i){
	  boundaryHeap[inputView[i][boundary[i]]] = i;
	}
	it = boundaryHeap.end();
	--it;
	unsigned int numToReduceBy = currentPartitionSize - currentSizeGoal;
	for(unsigned int i=0;i<numToReduceBy;++i){
	  unsigned int maxIndex =(*it).second; 
	  boundaryHeap.erase(it);
	  if (boundary[i]-step < 0 ){
	    boundary[i] = -1;
	  } else {
	    boundary[maxIndex] -= step;
	  }
	  if (boundary[i] != -1) {
	    boundaryHeap[inputView[maxIndex][boundary[maxIndex]]] = maxIndex;
	  }
	  it = boundaryHeap.end();
	  --it;
	}
	//update lMax
	lMax = (*it).first;
      }
      //increase sample for next iteration
      step = step /2;
      currentSizeGoal = static_cast<size_t>(floor(partitionFraction * floor(inputView[0].size() / static_cast<double>(step))*inputView.size()));
    }
    return boundary;
  }

  //merge the views in input from boundary (inclusive) untill numToMerge elements have been merged.
  template<class Boundary, class InputView, class OutputIter, class Compare>
  void MultiWayMerge(InputView& input, Boundary& boundary, OutputIter& outIter, size_t numToMerge, Compare cmp){
    typename std::map<typename OutputIter::value_type, unsigned int, Compare> mergeHeap(cmp);
    for(size_t i=0;i<input.size();++i){
      if(boundary[i]>=0 && boundary[i] < input[i].size()){
	mergeHeap[input[i][boundary[i]]] = i;
      }
    }
    typename std::map<typename OutputIter::value_type,unsigned int, Compare>::iterator inIter = mergeHeap.begin();
    for(size_t numMerged = 0;numMerged<numToMerge;++numMerged) {
      //write the smallest value to output;
      *outIter =(*inIter).first;
      ++outIter;
      //insert the next value from the list
      unsigned int index = (*inIter).second;
      mergeHeap.erase(inIter);
      ++boundary[index];
      if(boundary[index] != (input[index]).size()){
	mergeHeap[input[index][boundary[index]]] = index;
      }
      inIter = mergeHeap.begin();
    }
  }

  //note: that topview must be Topview<SubView<BaseView>>
  template<class TopView, class BaseView, class Compare, class Task>
  class SubMultiWayMergeTask : public Task {
  public:
    SubMultiWayMergeTask(TopView* inputView, BaseView* outputview, size_t numToMerge, size_t myTaskNum, Compare cmp)
      : inputView_(inputView), outputView_(outputview), numToMerge_(numToMerge), myTaskNum_(myTaskNum), cmp_(cmp) {
    }

    void operator()() {
      BasicArray<size_t> partitionBoundary(inputView_->size());
      if (myTaskNum_ == 0){
	for(size_t i=0;i<inputView_->size();++i) {
	  partitionBoundary[i] = 0;
	}
      }
      if (myTaskNum_ != 0){
	BasicArray<long int> prePartitionBoundary(inputView_->size());
	prePartitionBoundary = FWayPartition(numToMerge_ * myTaskNum_, *inputView_,cmp_);
	//increment the boundary to get the indicies the this task should start to work from
	for(size_t i=0;i<inputView_->size();++i){
	  partitionBoundary[i] = ++prePartitionBoundary[i];
	}
      }
      typename BaseView::iterator outIt = outputView_->begin();
      outIt += numToMerge_*myTaskNum_;
      MultiWayMerge(*inputView_,partitionBoundary,outIt,numToMerge_,cmp_);
    }
  private:
    TopView* inputView_;
    BaseView* outputView_;
    size_t numToMerge_;
    size_t myTaskNum_;
    Compare cmp_;
  };

  template<class TopView, class BaseView, class Compare, class Task>
  class MultiWayMergeTask : public Task {
  public:
    MultiWayMergeTask(TopView* inputView, BaseView* outputView, size_t numTasks, Compare cmp)
      : inputView_(inputView), outputView_(outputView), numTasks_(numTasks), cmp_(cmp) {}
      
    void operator()(){
      size_t numToMerge = outputView_->size()/numTasks_;
      typedef MultiWayMergeTask<TopView, BaseView, Compare, Task> MergeTask;
      BasicArray<Task*> mergeTasks(numTasks_);
      for(size_t i=0;i<numTasks_;++i){
	mergeTasks[i] = new SubMultiWayMergeTask<TopView, BaseView, Compare, Task>(inputView_,outputView_,numToMerge,i,cmp_);
      }
      Spawn(mergeTasks);
      for(size_t i=0;i<numTasks_;++i){
	delete mergeTasks[i];
      }
    }
  private:
    TopView* inputView_;
    BaseView* outputView_;
    size_t numTasks_;
    Compare cmp_;
  };

}//end namespace mcl
#endif //header guard

