#ifndef MCL_ATOMIC_ARRAY
#define MCL_ATOMIC_ARRAY

#include <initializer_list>
#include <algorithm>
#include <stdexcept>
#include <iterator>

#include <tbb/atomic.h>

namespace mcl {

  /**
   * The atomic iterator does NOT allow concurrent usage, but is used to iterate a atomic container.
   * Each task/thread should use its own iterator.
   **/
  template<class T>
  class AtomicArrayIterator : public std::iterator<std::random_access_iterator_tag,tbb::atomic<T>*>{
  public:
    typedef typename std::iterator<std::random_access_iterator_tag,tbb::atomic<T>>::value_type value_type;
    typedef typename std::iterator<std::random_access_iterator_tag,tbb::atomic<T>>::reference reference;
    typedef typename std::iterator<std::random_access_iterator_tag,tbb::atomic<T>>::difference_type difference_type;
    typedef typename std::iterator<std::random_access_iterator_tag,tbb::atomic<T>>::pointer pointer;
    typedef typename std::iterator<std::random_access_iterator_tag,tbb::atomic<T>>::iterator_category iterator_category;

    inline AtomicArrayIterator() : pos_(0) {}
    inline AtomicArrayIterator(pointer pos) : pos_(pos) {}
    inline AtomicArrayIterator(reference pos) : pos_(pos) {}
    inline AtomicArrayIterator(const AtomicArrayIterator<T>& rhs) : pos_(rhs.pos_) {}
    inline AtomicArrayIterator<T>&  operator=(const AtomicArrayIterator<T>& rhs)
      {
	AtomicArrayIterator<T> tmp(rhs);
	this->swap(tmp);
	return *this;
      }
    inline void swap(AtomicArrayIterator<T>& rhs) {std::swap(this->pos_,rhs.pos_);}
    inline bool operator==(const AtomicArrayIterator<T>& rhs) const {return pos_ == rhs.pos_;}
    inline bool operator!=(const AtomicArrayIterator<T>& rhs) const {return !this->operator==(rhs);}
    inline bool operator<(const AtomicArrayIterator<T>& rhs) const {return pos_ < rhs.pos_;} //operator< and > workes for different arrays
    inline bool operator>(const AtomicArrayIterator<T>& rhs) const {return rhs.operator<(this);}
    inline bool operator<=(const AtomicArrayIterator<T>& rhs) const {return (this->operator<(rhs)||this->operator==(rhs));}
    inline bool operator>=(const AtomicArrayIterator<T>& rhs) const {return (this->operator>(rhs)||this->operator==(rhs));}
    inline reference operator*() {return *pos_;}
    inline reference operator->() {return *pos_;}
    inline reference operator[](const difference_type& n) {return pos_[n];}
    inline AtomicArrayIterator<T>& operator--() {--pos_; return *this;}//prefix
    inline AtomicArrayIterator<T> operator--(int) {AtomicArrayIterator<T> tmp(*this); --pos_; return tmp;}//postfix
    inline AtomicArrayIterator<T>& operator++(){++pos_; return *this;}//prefix
    inline AtomicArrayIterator<T> operator++(int) {AtomicArrayIterator<T> tmp(*this); ++pos_; return tmp;}//postfix
    inline AtomicArrayIterator<T> operator+(const difference_type& n) const {return AtomicArrayIterator<T>(pos_ + n);}
    inline AtomicArrayIterator<T> operator-(const difference_type& n) const {return AtomicArrayIterator<T>(pos_ + n);}
    inline difference_type operator-(const AtomicArrayIterator<T>& rhs) const {return pos_ - rhs.pos_;}
    inline AtomicArrayIterator<T>& operator+=(const difference_type& n) {pos_ += n; return *this;}
    inline AtomicArrayIterator<T>& operator-=(const difference_type& n) {pos_ -= n; return *this;}

  protected: 
    tbb::atomic<T>* pos_;
 };
  

  /**
   * TO DO: Make changing the array atomic 
   *           i.e. copy construction, assignment, and swap
   *
   *        This require a memory mangement system eg hazard pionters.
   *        Apply hazard pointers
   *        Construct handle class, to allow thread mangement and HZP
   *
   * Allocation, deallocation assignment swap and copy construction 
   *   are NOT atomic and when used while other threads are using the structure the behaivour is undefined.
   *  
   **/

  template<class T> 
    class AtomicArray {
  public:
    //view trait types
    typedef T value_type;
    typedef tbb::atomic<T> atomic_value_type;
    typedef void key_type;
    typedef size_t size_type;
    typedef tbb::atomic<T>& data_type;
    typedef size_t index_type;
    typedef tbb::atomic<T>& reference;
    typedef const tbb::atomic<T>& const_reference;
    typedef tbb::atomic<T> pointer;
    typedef const tbb::atomic<T> const_pointer;
    typedef AtomicArrayIterator<T> iterator;
    typedef AtomicArrayIterator<const T> const_iterator;
    
    inline explicit AtomicArray(size_t n) : length_(n) {
      data_ = new tbb::atomic<T>[length_+1]; //should the array be initialized?
    }
    inline AtomicArray(AtomicArray<T>& rhs) : length_(rhs.size()) {
      data_ = new tbb::atomic<T>[length_+1];
      std::copy(&rhs.data_[0],&rhs.data_[length_],&data_[0]);
    }
    inline AtomicArray(std::initializer_list<T> list) : length_(list.size()) {
      data_ = new tbb::atomic<T>[length_+1];
      std::copy(list.begin(),list.end(),data_);
    }
    inline ~AtomicArray(){
      delete[] data_;
    }
    inline void swap(AtomicArray<T>& rhs){
      std::swap(this->length_,rhs.length_);
      std::swap(this->data_,rhs.data_);
    }
    inline AtomicArray<T>& operator=(const AtomicArray<T>& rhs){
      AtomicArray<T> tmp(rhs);
      this->swap(tmp);
      return this;
    }
    inline data_type read(const index_type& n) const {
      if(n >length_ -1)
	throw std::out_of_range("index is too large");
      return data_[n];
    }
    inline void write(const index_type& n, const value_type& val){
      if(n >length_ -1)
	throw std::out_of_range("index is too large");
      data_[n] = val;
    }
    inline value_type&& try_write(const index_type& n,
				     const value_type& old,
				     const value_type& val){
      if (n > length_ -1)
	throw std::out_of_range("index is too large");
      return std::forward<value_type>(data_[n].compare_and_swap(val,old));
    }
    inline data_type operator[](const index_type& n){
      return data_[n];
    }
    inline const data_type operator[](const index_type& n) const{
      return data_[n];
    }
    inline data_type at(const index_type& n){
      if(n >length_ -1)
	throw std::out_of_range("index is too large");
      return data_[n];
    }
    inline const data_type at(const index_type n) const{
      if(n >length_ -1)
	throw std::out_of_range("index is too large");
      return data_[n];
    }
    inline bool operator==(const AtomicArray<T>& rhs) const {
      return (length_ == rhs.size() && std::equal(data_[0],data_[length_],rhs.data_[0]));
    }
    inline bool operator!=(const AtomicArray<T>& rhs) const {
      return (length_ != rhs.size()|| !std::equal(data_[0],data_[length_],rhs.data_[0]));
    }
    inline size_type size() const {
      return length_;
    }
    inline bool empty() const {
      return length_ == 0; //actual used elements?
    }
    inline iterator begin() {
      return iterator(data_);
    }
    inline const_iterator begin() const {
      return const_iterator(data_);
    }
    inline iterator end() {
      return iterator(data_ + length_);
    }
    inline const_iterator end() const {
      return const_iterator(data_ + length_ + 1);
    }
    inline reference front() {
      return data_[0];
    }
    inline const_reference front() const {
      return data_[0];
    }
    inline reference back() {
      return data_[length_];
    }
    inline const_reference back() const {
      return data_[length_];
    }
  protected:
    tbb::atomic<T>* data_;
    size_t length_;
  };

}//end namespace mcl

#endif //header gard
