#ifndef MCL_BASIC_ARRAY
#define MCL_BASIC_ARRAY

#include <iostream>

#include <initializer_list>
#include <algorithm>
#include <stdexcept>
#include <iterator>

namespace mcl {
 
  template<class T>
    class ArrayIterator : public std::iterator<std::random_access_iterator_tag,T*> {
  public:
    //typedefs
    typedef typename std::iterator<std::random_access_iterator_tag, T>::value_type value_type;
    typedef typename std::iterator<std::random_access_iterator_tag, T>::reference reference;
    typedef typename std::iterator<std::random_access_iterator_tag, T>::difference_type difference_type;
    typedef typename std::iterator<std::random_access_iterator_tag, T>::pointer pointer;

    inline ArrayIterator() : pos_(0) {}
    inline ArrayIterator(pointer pos) : pos_(pos) { }
    inline ArrayIterator(reference pos) : pos_(pos) {}
    inline ArrayIterator(const ArrayIterator<T>& rhs) : pos_(rhs.pos_) {}
    inline ArrayIterator<T>&  operator=(const ArrayIterator<T>& rhs)
      {
	ArrayIterator<T> tmp(rhs);
	this->swap(tmp);
	return *this;
      }
    inline void swap(ArrayIterator<T>& rhs) {std::swap(this->pos_,rhs.pos_);}
    inline bool operator==(const ArrayIterator<T>& rhs) const {
      //std::cout<<"BasicArray::iterator::=="<<pos_<<" == "<<rhs.pos_<<std::endl;
      return pos_ == rhs.pos_;}
    inline bool operator!=(const ArrayIterator<T>& rhs) const {return !this->operator==(rhs);}
    inline bool operator<(const ArrayIterator<T>& rhs) const {return pos_ < rhs.pos_;} //operator< and > workes for different arrays
    inline bool operator>(const ArrayIterator<T>& rhs) const {return !operator<(rhs);}
    inline bool operator<=(const ArrayIterator<T>& rhs) const {return (this->operator<(rhs)||this->operator==(rhs));}
    inline bool operator>=(const ArrayIterator<T>& rhs) const {return (this->operator>(rhs)||this->operator==(rhs));}
    inline reference operator*() {return *pos_;}
    inline reference operator->() {return *pos_;}
    inline reference operator[](const difference_type& n) {return pos_[n];}
    inline ArrayIterator<T>& operator--() {--pos_; return *this;}//prefix
    inline ArrayIterator<T> operator--(int) {ArrayIterator<T> tmp(*this); --pos_; return tmp;}//postfix
    inline ArrayIterator<T>& operator++(){++pos_; return *this;}//prefix
    inline ArrayIterator<T> operator++(int) {ArrayIterator<T> tmp(*this); ++pos_; return tmp;}//postfix
    inline ArrayIterator<T> operator+(const difference_type& n) const {return ArrayIterator<T>(pos_ + n);}
    inline ArrayIterator<T> operator-(const difference_type& n) const {return ArrayIterator<T>(pos_ - n);}
    inline difference_type operator-(const ArrayIterator<T>& rhs) const {return pos_ - rhs.pos_;}
    inline ArrayIterator<T>& operator+=(const difference_type& n) {pos_ += n; return *this;}
    inline ArrayIterator<T>& operator-=(const difference_type& n) {pos_ -= n; return *this;}
    friend std::ostream& operator<<(std::ostream& out, const ArrayIterator<T>& it) {
      it.Print(out);
      return out;
    }
  protected:
    T* pos_;
    void Print(std::ostream& out) const {
      out<<"ArrayIterator pos="<<pos_;
    }
  };

  /**
   * TO DO: Make changing the array atomic 
   *           i.e. copy construction, assignment, and swap
   *
   *        This require a memory mangement system eg hazard pionters.
   *        Apply hazard pointers
   *        Construct handle class, to allow thread mangement and HZP
   *
   * Allocation, deallocation assignment swap and copy construction 
   *   are NOT atomic and when used while other threads are using the structure the behaivour is undefined.
   *  
   **/

  template<class T>
    class BasicArray {
  public:
    //view trait types
    typedef T value_type;
    typedef void key_type; 
    typedef size_t size_type;
    typedef T& data_type;
    typedef size_t index_type;
    typedef T& reference;
    typedef const T& const_reference;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef ArrayIterator<T> iterator;
    typedef ArrayIterator<const T> const_iterator;

    explicit inline BasicArray(const size_type& n=0) : length_(n)
    {
      //std::cout<<"constructing BasicArray with size"<<n<<std::endl;
      array_ = new T [length_+1];//reserve space for a past the end iterator
    }
    inline BasicArray(const BasicArray<T>& rhs) : length_(rhs.size()) {
      array_ = new T [length_];
      std::copy(&rhs.array_[0],&rhs.array_[length_],&array_[0]);
    }
    inline BasicArray(std::initializer_list<T> list) : length_(list.size()){
      //std::cout<<"constructing BasicArray from list, length = "<<length_<<std::endl;
      array_ = new T [length_];
      std::copy(list.begin(),list.end(),array_);
    }
    inline ~BasicArray(){
      delete[] array_;
      //std::cout<<"BasicArray "<<this<<" dead!"<<std::endl;
    }
    inline BasicArray<T>& operator=(const BasicArray<T>& rhs){
      BasicArray<T> tmp(rhs);
      swap(tmp);
      return *this;
    }
    inline void swap(BasicArray<T>& rhs){
      std::swap(array_,rhs.array_);
      std::swap(length_,rhs.length_);
    }
    inline data_type operator[](const index_type& n) {
      return array_[n];
    }
    inline const data_type operator[](const index_type& n) const {
      return array_[n];
    }
    inline data_type at(const index_type& n) { //bound checked
      if (n > length_-1)
	throw std::out_of_range("index is too large");
      return array_[n];
    }
    inline const data_type at(const index_type& n) const {
      if (n > length_-1)
	throw std::out_of_range("index is too large");
      return array_[n];
    }
    inline size_type size() const {
      return length_;
    }
    inline bool empty() const {
      return length_ == 0;
    }
    inline iterator begin() {
      return iterator(array_);
    }
    inline const_iterator begin() const {
      return const_iterator(array_);
    }
    inline iterator end() {
      return iterator(array_ + length_);
    }
    inline const_iterator end() const {
      return const_iterator(array_ + length_ + 1);
    }
    inline data_type& front() {
      return array_[0];
    }
    inline const data_type& front() const {
      return array_[0];
    }
    inline data_type back() {
      return array_[length_];
    }
    inline const data_type back() const {
      return array_[length_];
    }
    inline bool operator==(const BasicArray<T>& rhs) const {
      return (length_ == rhs.size() && std::equal(array_,array_+length_,rhs.array_));
    }
    inline bool operator!=(const BasicArray<T>& rhs) const {
      return (length_ != rhs.size() || !std::equal(array_,array_+length_,rhs.array_));
    }
    inline data_type read(const index_type& n) const {
      if (n > length_ -1)
	throw std::out_of_range("index is too large");
      return array_[n];
    }
    inline void write(const index_type& n, data_type val){
      at(n) = val;
    }

  protected:
    T* array_; //one larger than length_, array_+length_ is reserved for the past the end iterator
    size_t length_;
  };




  
}//end of namespace mcl

#endif //include garde


