#ifndef MCL_PARALLEL_SCAN_VIEW
#define MCL_PARALLEL_SCAN_VIEW

#include <stdexcept>

#include "MCL/data_structures/atomicArray.h"


namespace mcl {

  template<class BaseView>
  class ParallelScanView;

  template<class BaseIter, class ScanView>
  class ParallelScanIterator {
  public:
    typedef typename BaseIter::value_type value_type;
    typedef typename BaseIter::reference reference;
    typedef typename BaseIter::difference_type difference_type;
    typedef typename BaseIter::pointer pointer;
    typedef typename ScanView::atomicArray_type atomicArray_type;

    inline ParallelScanIterator() : basePos_(), atomicPos_() {}

    inline ParallelScanIterator(BaseIter base, typename atomicArray_type::iterator atomic) 
      : basePos_(base), atomicPos_(atomic) {}
    
    inline ParallelScanIterator(const ParallelScanIterator<BaseIter, ScanView>& rhs)
      : basePos_(rhs.basePos_), atomicPos_(rhs.atomicPos_) {}

    inline ParallelScanIterator<BaseIter, ScanView>& 
      operator=(const ParallelScanIterator<BaseIter, ScanView>& rhs) {
      basePos_ = rhs.basePos_;
      atomicPos_ = rhs.atomicPos_;
      return *this;
    }

    inline void swap(ParallelScanIterator<BaseIter, ScanView>& rhs) {
      std::swap(basePos_,rhs.basePos_);
      std::swap(atomicPos_,rhs.atomicPos_);
    }
    
    inline bool operator==(const ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return (basePos_ == rhs.basePos_);
    }

    inline bool operator!=(const ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return !operator==(rhs);
    }

    inline bool operator<(const ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return basePos_ < rhs.basePos_;
    }

    inline bool operator>(const ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return basePos_ > rhs.basePos_;
    }

    inline bool operator<=(const ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return (operator<(rhs) || operator==(rhs));
    }

    inline bool operator>=(const ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return (operator>(rhs) || operator==(rhs));
    }
    inline reference operator*() {return *basePos_;}
    inline reference operator->() {return *basePos_;}
    inline reference operator[](const difference_type& n) {return basePos_[n];}
    inline ParallelScanIterator<BaseIter, ScanView>& operator--() {
      --basePos_;
      --atomicPos_;
      return *this; 
    }
    inline ParallelScanIterator<BaseIter, ScanView>& operator++() {
      ++basePos_;
      ++atomicPos_;
      return *this;
    }
    inline ParallelScanIterator<BaseIter, ScanView> operator--(int) {
      ParallelScanIterator<BaseIter, ScanView> ans(this);
      --basePos_;
      --atomicPos_;
      return ans;
    }
    inline ParallelScanIterator<BaseIter, ScanView> operator++(int) {
      ParallelScanIterator<BaseIter, ScanView> ans(this);
      ++basePos_;
      ++atomicPos_;
      return ans;
    }
    inline difference_type operator-(ParallelScanIterator<BaseIter, ScanView>& rhs) const {
      return (basePos_ - rhs.basePos_);
    }
    inline ParallelScanIterator<BaseIter, ScanView> operator+(const difference_type& n) const {
      ParallelScanIterator<BaseIter, ScanView> ans(*this);
      ans.basePos_ += n;
      ans.atomicPos_ += n;
      return ans;
    }
    inline ParallelScanIterator<BaseIter, ScanView> operator-(const difference_type& n) const {
      ParallelScanIterator<BaseIter, ScanView> ans(*this);
      ans.basePos_ -= n;
      ans.atomicPos_ -= n;
      return ans;
    }
    inline ParallelScanIterator<BaseIter, ScanView>& operator+=(const difference_type& n){
      basePos_ += n;
      atomicPos_ += n;
    }
    inline ParallelScanIterator<BaseIter, ScanView>& operator-=(const difference_type& n){
      basePos_ -= n;
      atomicPos_ -= n;
    }
    //increments positions untill a position where the flag has not been set is found
    // or the next position is stop.
    //(stop is past the end)
    // if no positions are awailable after stop the a range_error exception is thrown
    // note: stop > this
    inline ParallelScanIterator<BaseIter, ScanView>& 
    GetNextBefore(const ParallelScanIterator<BaseIter, ScanView>& stop) {
      while(atomicPos_ != stop.atomicPos_){
	if (!(*atomicPos_)){
	  if (!((*atomicPos_).compare_and_swap(true,false))) {
	    return *this;
	  }
	}
	operator++();
      }
      throw std::range_error("no available positions");
    }
    
    //decrements positions untill a position where the flag has not been set is found
    // or the stop position is reached.
    //(stop is included)
    // if no positions are awailable before stop the a range_error exception is thrown
    // note: stop < this
    // note: if called on end end can be returned
    inline ParallelScanIterator<BaseIter, ScanView>& 
    GetPrevAfter(const ParallelScanIterator<BaseIter, ScanView>& stop) {
      //std::cout<<"GetPrevAfter"<<std::endl;
      while(atomicPos_ != stop.atomicPos_){
	if (!(*atomicPos_)){
	  if (!((*atomicPos_).compare_and_swap(true,false))) {
	    return *this;
	  }
	}
	operator--();
      }
      throw std::range_error("no available positions");
    }

  protected:
    BaseIter basePos_;
    typename atomicArray_type::iterator atomicPos_;
  };


  template<class BaseView>
  class ParallelScanView {
  public:
    typedef typename BaseView::value_type value_type;
    typedef typename BaseView::key_type key_type;
    typedef typename BaseView::size_type size_type;
    typedef typename BaseView::data_type data_type;
    typedef typename BaseView::index_type index_type;
    typedef typename BaseView::reference reference;
    typedef typename BaseView::const_reference const_reference;
    typedef typename BaseView::pointer pointer;
    typedef typename BaseView::const_pointer const_pointer;
    typedef ParallelScanIterator<typename BaseView::iterator, ParallelScanView<BaseView> > iterator;
    typedef ParallelScanIterator<typename BaseView::const_iterator, ParallelScanView<BaseView> > const_iterator;
    typedef AtomicArray<bool> atomicArray_type;
    
    explicit inline ParallelScanView(BaseView* base) 
      : flags_(base->size()), base_(base) {
      //std::cout<<"ParallelScanView(base): initializing atomic array."<<std::endl;
      AtomicArray<bool>::iterator it = flags_.begin();
      while(it != flags_.end()){
	*it = false;
	++it;
      }
    }

    inline ParallelScanView(const ParallelScanView<BaseView>& rhs) 
      : flags_(rhs.flags_), base_(rhs.base_) {}
   
    inline ParallelScanView<BaseView>& operator=(const ParallelScanView<BaseView>& rhs) {
      flags_(rhs.flags_);
      base_(rhs.base_);
      //std::cout<<"parllelScanView assignment!!"<<std::endl;
      return *this;
    }
    
    inline void swap(ParallelScanView<BaseView>& rhs){
      std::swap(flags_,rhs.flags_);
      std::swap(base_,rhs.base_);
    }
    
    inline data_type operator[](const index_type& n){
      return (*base_)[n];
    }
    
    inline const data_type operator[](const index_type& n) const {
      return (*base_)[n];
    }

    inline data_type at(const index_type& n){
      return base_->at(n);
    }

    inline const data_type at(const index_type& n) const {
      return base_->at(n);
    }

    inline size_type size() const {
      return base_->size();
    }

    inline bool empty() const {
      return base_->empty();
    }

    inline bool operator==(const ParallelScanView<BaseView>& rhs) const {
      return (flags_ == rhs.flags_ && base_ == rhs.base_);
    }

    inline bool operator!=(const ParallelScanView<BaseView>& rhs) const {
      return !operator==(rhs);
    }

    inline iterator begin(){
      return iterator(base_->begin(),flags_.begin());
    }
    inline const_iterator begin() const {
      return const_iterator(base_->begin(),flags_.begin());
    }
    inline iterator end() {
      return iterator(base_->end(),flags_.end());
    }
    inline const_iterator end() const {
      return const_iterator(base_->end(),flags_.end());
    }
    friend std::ostream& operator<<(std::ostream& out, const ParallelScanView<BaseView>& sv) {
      sv.Print(out);
      return out;
    }
  protected:
    atomicArray_type flags_;
    BaseView* base_;
    void Print(std::ostream& out) const{
      out<<"ScanView(base:"<<(*base_)<<")";
    }
  };

}//end namespace mcl
#endif //header guard
