#ifndef MCL_SUBVIEW
#define MCL_SUBVIEW

#include <stdexcept>
#include <cassert>

namespace mcl {

  template<class BaseView>
  class SubView {
  public:
    //view trait types
    typedef typename BaseView::value_type value_type;
    typedef typename BaseView::key_type key_type; 
    typedef typename BaseView::size_type size_type;
    typedef typename BaseView::data_type data_type;
    typedef typename BaseView::index_type index_type;
    typedef typename BaseView::reference reference;
    typedef typename BaseView::const_reference const_reference;
    typedef typename BaseView::pointer pointer;
    typedef typename BaseView::const_pointer const_pointer;
    typedef typename BaseView::iterator iterator;
    typedef typename BaseView::const_iterator const_iterator;
    typedef BaseView BaseView_type;

    inline SubView() : base_(), start_(0), end_(0) {
      //std::cout<<"empty suview:"<<this<<std::endl; 
    }

    inline SubView(BaseView* b, index_type s, index_type e) 
      : base_(b), start_(s), end_(e) {
      //std::cout<<"SubView( base_"<<base_<<") "<<this<<std::endl; 
    }

    inline SubView(SubView<BaseView>* b, index_type s, index_type e) : 
      base_(b->base_), start_(b->start_+s), end_(b->start_+e) {
      //std::cout<<"SubView( base_"<<base_<<") "<<this<<std::endl; 
    }

    inline SubView(SubView<BaseView>& b, index_type s, index_type e) : 
      base_(b.base_), start_(b.start_+s), end_(b.start_+e) {
      //std::cout<<"SubView( base_"<<base_<<") "<<this<<std::endl;
    }

    inline SubView(const SubView<BaseView>& rhs)
      : base_(rhs.base_), start_(rhs.start_), end_(rhs.end_){ 
      //std::cout<<"SubView "<<this<<" copy base"<<base_<<std::endl;
      //std::cout<<"Subview copy from: "<<rhs<<std::endl;
    }

    inline ~SubView(){
      //std::cout<<"SubView "<<this<<" dieing"<<std::endl;
    }

    inline void swap(SubView<BaseView>& rhs){
      std::swap(this->base_,rhs.base_);
      std::swap(this->start_,rhs.start);
      std::swap(this->end_,rhs.end_);
    }
    inline void swapElements(SubView<BaseView>& rhs){
      if (rhs.size() != size()){
	std::cout<<"this->size() = "<<size()<<" rhs.size()="<<rhs.size()<<std::endl;
	throw std::logic_error("SubView sizes do not match");
      }
      typename BaseView::iterator lhsIter = begin();
      typename BaseView::iterator rhsIter = rhs.begin();
      while(lhsIter != end()){
	std::swap(*lhsIter,*rhsIter);
	++lhsIter;
	++rhsIter;
      }
    }
    inline void swapElements(SubView<BaseView> rhs){
      if (rhs.size() != size()){
	std::cout<<"this->size() = "<<size()<<" rhs.size()="<<rhs.size()<<std::endl;
	throw std::logic_error("SubView sizes do not match");
      }
      typename BaseView::iterator lhsIter = begin();
      typename BaseView::iterator rhsIter = rhs.begin();
      while(lhsIter != end()){
	std::swap(*lhsIter,*rhsIter);
	++lhsIter;
	++rhsIter;
      }
    }
    inline SubView<BaseView>& operator=(const SubView<BaseView>& rhs){
      base_ = rhs.base_;
      start_ = rhs.start_;
      end_ = rhs.end_;
      //std::cout<<"SubView "<<this<<" assignment base"<<base_<<std::endl;
      return *this;
    }
    inline size_type size() const {
      //std::cout<<"subView::size(start:"<<start_<<" end:"<<end_<<") size is:"<<end_ - start_<<std::endl;
      return end_ - start_; 
    }
    inline bool empty(){
      return size()==0;
    }
    inline data_type operator[](const index_type& n){
      assert(base_ != NULL);
      return (*base_)[n+start_];
    }
    inline const data_type operator[](const index_type& n) const {
      assert(base_ != 0);
      return (*base_)[n+start_];
    }
    inline data_type at(const index_type& n){
      if (n > (end_ - start_ +1))
	throw std::out_of_range("index too large");
      return base_->at(n+start_);
    }
    inline const data_type at(const index_type& n) const{
      if (n > (end_ - start_))
	throw std::out_of_range("index too large");
      return base_->at(n+start_);
    }
    inline iterator begin(){
      return base_->begin()+start_;
    }
    inline const_iterator begin() const{
       return base_->begin()+start_;
    }
    inline iterator end(){
      return base_->begin()+end_;
    }
    inline const_iterator end() const{
      return base_->begin()+end_;
    }
    inline data_type front() {
      return base_[start_];
    }
    inline const data_type front() const {
      return (*base_)[start_];
    }
    inline data_type back() {
      return (*base_)[end_];
    }
    inline const data_type back() const {
      return (*base_)[end_];
    }  
    inline bool operator==(const SubView<BaseView>& rhs) const {
      return (start_ == rhs.start_ && end_ == rhs.end_ && base_ == rhs.base_);
    }
    inline bool operator==(const BaseView& rhs){
      return (start_ == 0 && end_ == rhs.size() && base_ == rhs);
    }
    inline bool operator!=(const SubView<BaseView>& rhs) const {
      return (start_ != rhs.start_ || end_ != rhs.end_ || base_ != rhs.base_);
    }
    inline bool operator!=(const BaseView& rhs){
      return (start_ != 0 || end_ != rhs.size() || base_ != rhs);
    }
    inline data_type read(const index_type& n) const {
      if(n > end_ - start_)
    	throw std::out_of_range("index too large");
      return base_->at(std::forward<index_type>(n));
    }
    inline void write(const index_type& n, data_type val){
      base_->write(std::forward<index_type>(n),std::forward<data_type>(val));
    }
    friend std::ostream& operator<<(std::ostream& out, const SubView<BaseView>& sv) {
      sv.Print(out);
      return out;
    }
    // for debugging
    inline BaseView* Base(){
      return base_;
    }
  protected:
    BaseView* base_;
    index_type start_;
    index_type end_; //past the end
  private:
    void Print(std::ostream& out) const{
      out<<"SubView(base:"<<base_<<" start:"<<start_<<" end:"<<end_<<")";
    }
  };

}//end namespace mcl

#endif //header garde
