#ifndef MCL_TRANSFORM_VIEW
#define MCL_TRANSFORM_VIEW

#include <functional>

namespace mcl {

  /**
   * Because std::function objects are NOT equality combarable,
   *  are the equality comparator implemented without transformation function comparison
   **/
 
  template<class BaseIterator, class ResultType>
  class TransformIterator {
  public:
    typedef ResultType value_type;
    typedef ResultType& refernce;
    typedef typename BaseIterator::difference_type difference_type;
    typedef ResultType* pointer;
    typedef std::function<ResultType(typename BaseIterator::value_type)> TransformFunction;
    
    inline TransformIterator(BaseIterator baseIt,TransformFunction function)
      : pos_(baseIt), transformFunction_(function) {}
    inline TransformIterator(const TransformIterator<BaseIterator,ResultType>& rhs)
      : pos_(rhs.pos_), transformFunction_(rhs.transformFunction_) {}
    inline TransformIterator<BaseIterator,ResultType>& operator=(const TransformIterator<BaseIterator,ResultType>& rhs){
      TransformIterator<BaseIterator,ResultType>& tmp(rhs);
      swap(tmp);
    }
    inline void swap(TransformIterator<BaseIterator,ResultType>& rhs){
      std::swap(pos_,rhs.pos_);
      std::swap(transformFunction_,rhs.transformFunction_);
    }
    inline bool operator==(const TransformIterator<BaseIterator,ResultType>& rhs){
      return (pos_ == rhs.pos_);
    }
    inline bool operator!=(const TransformIterator<BaseIterator,ResultType>& rhs){
      return !operator==(rhs);
    }
    inline bool operator<(const TransformIterator<BaseIterator,ResultType>& rhs){
      return pos_ < rhs.pos_;
    }
    inline bool operator<=(const TransformIterator<BaseIterator,ResultType>& rhs){
      return (operator<(rhs) || operator==(rhs));
    }
    inline bool operator>(const TransformIterator<BaseIterator,ResultType>& rhs){
      return pos_ > rhs.pos_;
    }
    inline bool operator>=(const TransformIterator<BaseIterator,ResultType>& rhs){
      return (operator>(rhs) || operator==(rhs));
    }
    inline value_type operator*(){return transformFunction_(*pos_);}
    inline value_type operator->(){return transformFunction_(*pos_);}
    inline TransformIterator<BaseIterator,ResultType>&
    operator[](const difference_type& n){pos_[n]; return *this;}
    
    inline TransformIterator<BaseIterator,ResultType>&
    operator++(){++pos_;return *this;}
    
    inline TransformIterator<BaseIterator,ResultType>&
    operator--(){--pos_;return *this;}

    inline TransformIterator<BaseIterator,ResultType>
    operator++(int){
      TransformIterator<BaseIterator,ResultType> tmp(*this);
      ++pos_;
      return tmp;
    }
    inline TransformIterator<BaseIterator,ResultType>
    operator--(int){
      TransformIterator<BaseIterator,ResultType> tmp(*this);
      --pos_;
      return tmp;
    }
    inline TransformIterator<BaseIterator,ResultType>
    operator+(const difference_type& n) const {
      return TransformIterator<BaseIterator,ResultType>(pos_+n,transformFunction_);
    }
    inline TransformIterator<BaseIterator,ResultType>
    operator-(const difference_type& n) const {
      return TransformIterator<BaseIterator,ResultType>(pos_-n,transformFunction_);
    }
    inline TransformIterator<BaseIterator,ResultType>&
    operator+=(const difference_type& n){ pos_ += n; return *this;}
    
    inline TransformIterator<BaseIterator,ResultType>&
    operator-=(const difference_type& n){ pos_ -= n; return *this;}
    
    friend std::ostream& operator<<(std::ostream& out, const TransformIterator<BaseIterator,ResultType>& it){
      it.Print(out);
      return out;
    }
  protected:
    void Print(std::ostream& out) const {
      out<<"transformIterator("<<pos_<<")";
    }
    BaseIterator pos_;
    TransformFunction transformFunction_;
  };

  /**
   * declaration of operator[] and at() is tricky as
   *  data_type& cannot be return directly
   **/
  template<class BaseView, class ResultType>
  class TransformView {
  public:
    typedef ResultType value_type;
    typedef typename BaseView::key_type key_type;
    typedef typename BaseView::size_type size_type;
    typedef ResultType& data_type;
    typedef typename BaseView::index_type index_type;
    typedef ResultType& reference;
    typedef const ResultType& const_reference;
    typedef ResultType* pointer;
    typedef const ResultType* const_pointer;
    typedef TransformIterator<typename BaseView::iterator,ResultType> iterator;
    typedef std::function<ResultType(typename BaseView::value_type)> TransformFunction;

    inline TransformView() : baseView_(0), transformFunction_() {}
    inline TransformView(BaseView* baseView,TransformFunction function)
      : baseView_(baseView), transformFunction_(function) {}
    inline TransformView(const TransformView<BaseView,ResultType>& rhs)
      : baseView_(rhs.baseView_), transformFunction_(rhs.transformFunction_){}
    inline TransformView<BaseView,ResultType>& 
    operator=(const TransformView<BaseView,ResultType>& rhs) {
      TransformView<BaseView,ResultType> tmp(rhs);
      swap(tmp);
      return *this;
    }
    inline void swap(TransformView<BaseView,ResultType>& rhs) {
      std::swap(baseView_,rhs.baseView_);
      std::swap(transformFunction_,rhs.transformFunction_);
    }
    inline size_type size(){
      return baseView_->size();
    }
    inline bool empty(){
      return baseView_->empty();
    }
    inline bool operator==(TransformView<BaseView,ResultType>& rhs){
      std::cout<<"transformView::iterator::=="<<std::endl;
      return (transformFunction_ == rhs.transformFunction
	      && *baseView_ == (*rhs.baseView_));
    }
    inline bool operator!=(TransformView<BaseView,ResultType>& rhs){
      return !operator==(rhs);
    }
        inline data_type operator[](const index_type& n){
      return transformFunction_(baseView_->operator[](n));
    }
    const data_type operator[](const index_type& n) const {
      return transformFunction_(baseView_->operator[](n));
    }
    inline data_type at(const index_type& n){
      return transformFunction(baseView_->at(n));
    }
    const data_type at(const index_type& n) const {
      return transformFunction(baseView_->at(n));
    }
    inline iterator begin(){
      return iterator(baseView_->begin(),transformFunction_);
    }
    // inline const_iterator begin() const{
    //   return const_iterator(baseView_->begin(),transformFunction_);
    // }
    inline iterator end(){
      return iterator(baseView_->end(),transformFunction_);
    }
    // inline const_iterator end() const{
    //   return const_iterator(baseView_->end(),transformFunction_);
    // }
    inline const data_type front() const {
      return transformFunction_(baseView_->front());
    }
    inline const data_type back() const {
      return transformFunction_(baseView_->back());
    }
  protected:
    BaseView* baseView_;
    TransformFunction transformFunction_;
  };

}//end namespace mcl

#endif //header guard
