#ifndef MCL_PDF_CENTRAL_LOCKING_TASK_QUEUE
#define MCL_PDF_CENTRAL_LOCKING_TASK_QUEUE

#include <algorithm>
#include <cassert>

#include <pthread.h>

#include "MCL/mutex.h"


namespace mcl {

  template<class Task>
  class LockingTaskQueue  {
    class Node;
  public:
    class Handle;
    
    //Queue
    inline LockingTaskQueue()
      : mutex_(), size_(0), numNotRunning_(0) {
      tail_ = new BasicNode(0,0,0);
      head_ = new BasicNode(0,0,0);
      head_->SetNext(tail_);
      head_->Running();
      tail_->Running();
    }
    inline ~LockingTaskQueue() {
      Mutex::Lock lock = mutex_.lock();
      Node* current = head_;
      Node* next = current->GetNext();
      while(next != tail_){
	delete current;
	current = next;
	next = next->GetNext();
      }
      delete next;
    }
    inline Handle GetHandle(Task* parent=0){
      Mutex::Lock lock = mutex_.lock();
      if(numNotRunning_ == 0 && size_ > 0){
	if(parent == 0 || !parent->IsReady()){
	  lock.Wait();
	}
      }
      Node* n = head_;
      while(n->IsRunning() && n != tail_){
	n = n->GetNext();
      }
      if(n != tail_){ 
	n = n->Running();
	--numNotRunning_;
	return Handle(n,this);
      }
      return Handle(NULL,this);
    }
    inline void AddTask(Task* task){
      Mutex::Lock lock = mutex_.lock();
      new BasicNode(task,tail_);
      ++size_;
      ++numNotRunning_;
      lock.Signal();
    }
    inline bool IsEmpty(){
      Mutex::Lock lock = mutex_.lock();
      return size_==0;
    }
  private:
    Mutex mutex_;
    Node* head_;
    Node* tail_;
    size_t size_; 
    size_t numNotRunning_;
    
    class Node {
    public:
      inline Node(Node* n, Node* p) : next_(n), prev_(p) {}
      inline Node() : next_(0), prev_(0) {}
      virtual inline ~Node() {}
      inline Node* GetNext(){
	return next_;
      }
      inline Node* GetPrev(){
	return prev_;
      }
      inline void SetNext(Node* n) {
	next_ = n;
	n->prev_ = this;
      }
      inline void SetPrev(Node* n){
	prev_ = n;
	n->next_ = this;
      }
      //given an Node n, insert n in the list just before this node
      inline void Insert(Node* p){
	p->next_ = this;
	if(prev_ != 0){
	  prev_->next_ = p;
	}
	p->prev_ = prev_;
	prev_ = p;
      }
      inline void Unlink(){
	prev_->next_ = next_;
	next_->prev_ = prev_;
      }
      virtual bool IsRunning(){return false;}
      virtual Node* Running(){return this;}
      virtual Task* GetTask(){return NULL;}
      virtual void SwapTask(Task* t){};
    protected:
      Node* next_;
      Node* prev_;
    };

    class BasicNode : public Node{
    public:
      inline BasicNode(Task* t, Node* n, Node* p)
	: Node(n,p), task_(t), running_(false) { }
      inline BasicNode(Task* t, Node* n) : Node(), task_(t), running_(false) {
	if(n != 0)
	  n->Insert(this);
      }
      inline void SwapTask(Task* t){
	task_ = t;
      }
      inline Node* Running(){
	running_ = true;
	return this ;
      }
      inline bool IsRunning(){
	return running_;
      }
      inline Task* GetTask(){
	return task_;
      }
    private:
      Task* task_;
      bool running_;
    };

    template<class TaskView>
    class ViewNode : public Node {
      typedef typename TaskView::iterator Iterator;
    public:
      inline ViewNode(Iterator first, Iterator last, Node* n, Node* p) 
	: Node(n,p), viewPos_(first), last_(last) {}
      inline ViewNode(Iterator first, Iterator last, Node* n)
	: Node(), viewPos_(first), last_(last){
	if (n != 0)
	  n->Insert(this);
      }
      inline bool IsRunning(){
	return false;
      }
      inline void SwapTask(Task* task){
	//not used but function is need by Node
      }
      inline Task* GetTask(){
	//not used but function needed by Node
	return *viewPos_;
      } 
      inline Node* Running(){
	BasicNode* n = new BasicNode(*viewPos_,this);
	n->Running();
	++viewPos_;
	if(viewPos_ == last_){
	  this->Unlink();
	  delete this;
	}
	return n;
      }   
    private:
      Iterator viewPos_;
      Iterator last_;
      //disable copyconstruction and assignment
      ViewNode(const ViewNode<TaskView>&);
      void operator=(const ViewNode<TaskView>&);
    };
    
  public:
    class Handle  {
    public:
      inline Handle(Node* n, LockingTaskQueue* q) 
	: queue_(q), currentPos_(n) {}
      inline Handle(const Handle& handle)
	:  queue_(handle.queue_), currentPos_(handle.currentPos_) {}
      inline Handle& operator=(const Handle& rhs) {
	currentPos_ = rhs.currentPos_;
	queue_ = rhs.queue_;
	return *this;
      }
      inline void swap(Handle& rhs){
	std::swap(currentPos_,rhs.currentPos_);
	std::swap(queue_,rhs.queue_);
      }
      template<class... Tasks>
      inline void InsertTasks(Tasks... tasks){
	Mutex::Lock lock = queue_->mutex_.lock();
	InternalInsert(currentPos_,tasks...);
	queue_->size_ += sizeof...(tasks);
	queue_->numNotRunning_ += sizeof...(tasks);
	//tell waiting threads that a new task is ready
	lock.Signal();
      } 
      template<class TaskView>
      inline void InsertView(TaskView& view){
	Mutex::Lock lock = queue_->mutex_.lock();
	new ViewNode<TaskView>(view.begin(), view.end(), currentPos_);
	queue_->size_ += view.size();
	queue_->numNotRunning_ += view.size();
	//tell waiting threads that new tasks are ready
	lock.Brodcast();
      }
      inline void Remove(){
	Mutex::Lock lock = queue_->mutex_.lock();
	currentPos_->Unlink();
	delete currentPos_;
	--queue_->size_;
	//tell waiting threads that their parent task might be ready
	lock.Brodcast();
      }
      inline Task* GetTask(){
	if (currentPos_ == 0 || currentPos_ == queue_->tail_ || currentPos_ == queue_->head_) {
	  return NULL;
	}
	return currentPos_->GetTask();
      }
    private:
      template<class... Tasks>
      Node* InternalInsert(Node* last){
	return last;
      }
      template<class... Tasks>
      Node* InternalInsert(Node* last, Task& head, Tasks... tail){
	BasicNode* next = InternalInsert(last,tail...);
	BasicNode* ans = new BasicNode(&head, next);
	return ans;
      }
      template<class... Tasks>
      Node* InternalInsert(Node* last, Task* head, Tasks... tail){
	BasicNode* ans = new BasicNode(head, InternalInsert(last,tail...));
	return ans;
      }
      LockingTaskQueue* queue_;
      Node* currentPos_;
    };
  };

}

#endif //header guard
