#ifndef MCL_WS_SCHEDULER
#define MCL_WS_SCHEDULER

#include <deque>
#include <iostream>
#include <tbb/atomic.h>

#include "MCL/core.h"
#include "MCL/mutex.h"
#include "MCL/data_structures/subview.h"

/**
 * TODO:
 *      -Comments
 *      -Inline small functions
 *      -Move inline functions to a inline.h file
 **/


namespace mcl {

  class LockingWorkStealingScheduler{
  public:
    class Handle;
    class Task;
    class TaskDeque;

    /*
     * Definition of the scheduler class:
     */
    inline LockingWorkStealingScheduler() {
      //seed the RNG used in work stealing
      srand(time(0));
    }
    inline ~LockingWorkStealingScheduler(){
      std::vector<Handle*>::iterator it = handles_.begin();
      while(it!=handles_.end()) {
	delete (*it);
      }
    }
    inline Handle* GetHandle() {
      Handle* h = new Handle(this);
      handles_.push_back(h);
      return h;
    }
    inline void AddTask(Task* task) {
      handles_.front()->AddTask(task);
    }
    inline bool IsDone(){
      std::vector<Handle*>::iterator it = handles_.begin();
      while(it!=handles_.end()) {
	if(!(*it)->IsEmpty()) {
	  return false;
	}
	++it;
      }
      return true;
    }
  private:
    inline Task* StealTask(){
      LockingWorkStealingScheduler::Handle* victim =
	handles_[rand() % handles_.size()];
      return victim->StealTask();
    }
    std::vector<Handle*> handles_;
    //disable copy construction and assignement
    LockingWorkStealingScheduler(const LockingWorkStealingScheduler&);
    void operator=(const LockingWorkStealingScheduler&);

  public:
    //MCL::LockingWorkStealingScheduler::Handle
    class Handle {
      friend class LockingWorkStealingScheduler;

    private:
      //MCL::LockingWorkStealingScheduler::Handle::TaskDeque
      class TaskDeque {
      public:
	TaskDeque() :  mutex_(), deque_() {}
	template<class... Tasks>
	inline void Push(Tasks... tasks){
	  Mutex::Lock lock = mutex_.lock();
	  InternalPush(tasks...);
	}
	//Pop: returns NULL if the deque is empty.
	inline Task* Pop() {
	  Mutex::Lock lock = mutex_.lock();
	  if (deque_.size() == 0) {
	    return NULL;
	  } else {
	    Task* task = deque_.front();
	    deque_.pop_front();
	    return task;
	  }
	}
	inline Task* StealTask() {
	  Mutex::Lock lock = mutex_.lock();
	  if(deque_.size() == 0) {
	    return NULL;
	  } else {
	    Task* task = deque_.back();
	    deque_.pop_back();
	    return task;
	  }
	}
	inline void AddTask(Task* task) {
	  Mutex::Lock lock = mutex_.lock();
	  deque_.push_back(task);
	}
	inline bool IsEmpty() {
	  Mutex::Lock lock = mutex_.lock();
	  size_t size = deque_.size();
	  return (size==0);
	}
      private:
	template<class... Tasks>
	inline void InternalPush() {}
	template<class... Tasks>
	inline void InternalPush(Task* task, Tasks... tasks){
	  InternalPush(tasks...);
	  deque_.push_front(task);
	}	
	Mutex mutex_;
	std::deque<Task*> deque_;
	//disallow copy construction and assignement
	TaskDeque(const TaskDeque&);
	void operator=(const TaskDeque&);
      };

      //Handle declaration
    public:
      explicit inline Handle(LockingWorkStealingScheduler* s) 
	: scheduler_(s), taskDeque_(), currentTask_(0), numRunningTasks_() {numRunningTasks_=0;}
      //implemented here to avoid impl.ementation file
      Task* TryGetTask(){
	Task* task = taskDeque_.Pop();
	if (task == NULL) {
	  //no local tasks try to steal one
	  Task* stolenTask = scheduler_->StealTask();
	  if (stolenTask != NULL){
	    currentTask_ = stolenTask;
	    stolenTask->SetSchedulerHandle(this);
	  }
	  return stolenTask;
	} else {
	  currentTask_ = task;
	  task->SetSchedulerHandle(this);
	  return task;
	}
      }
      //Spawning
      template<class... Tasks>
      void Spawn(Tasks... tasks){
	Task* parent = currentTask_;
	taskDeque_.Push(tasks...);
	BlockingRun(parent);
      }
      
      //used by spawn and task, calls scheduler_.IsDone();
      inline bool IsDone(){
	bool ans = scheduler_->IsDone();
	return ans;
      }
      inline void incRunningTasks(){
	++numRunningTasks_;
      }
      inline void decRunningTasks(){
	--numRunningTasks_;
      }
    private:
      inline void AddTask(Task* task){
	taskDeque_.AddTask(task);
      }
      inline Task* StealTask(){
	return taskDeque_.StealTask();
      }
      inline void BlockingRun(Task* parent){
	while(!parent->IsReady()) {
	  Task* task = TryGetTask();
	  if (task != NULL) {
	    task->Execute();
	    currentTask_ = parent;
	  }
	}
      }
      //used by scheduler IsDone()
      inline bool IsEmpty(){
	if(numRunningTasks_ == 0){ 
	  if (taskDeque_.IsEmpty()) {
	    if(numRunningTasks_ == 0)
	      return true;
	  }
	}
	return false;
      }
      LockingWorkStealingScheduler* scheduler_;
      TaskDeque taskDeque_;
      Task* currentTask_;
      tbb::atomic<int> numRunningTasks_;
      //disallow copy construction and assignement
      Handle(const Handle&);
      void operator=(const Handle&);
    };

    template<class View>
    class SplitTask;

    //mcl::LockingWorkStealingScheduler::Task.
    //defines mcl::Task when this scheduler is used.
    class Task {
    public:
      Task() : numSyncedChildren_(), numSpawnedChildren_(), parent_(0), schedulerHandle_(0) {}

      inline void Execute() {
	schedulerHandle_->incRunningTasks();
	operator()();
	schedulerHandle_->decRunningTasks();
	Task* parent = parent_;
	this->PostRun();
	if(parent != NULL) {
	  parent->IncNumSyncedChildren();
	}
      }
      virtual void operator()()=0;
      //PostRun function is called just before the execution of the task finishes
      //The task is NOT used after PostRun is called, making it possible to dealocate the task if needed
      virtual void PostRun() {}
      
      //allow spawing of a constant number of tasks
      //the tasks are inserted directly in the ready pool
      //NOTE: that tasks are added to the ready pool in reverse order!!
      template<class... Tasks>
      inline void Spawn(Tasks... tasks){
	SetParentsToThis(tasks...);
	numSpawnedChildren_ = sizeof...(tasks);
	schedulerHandle_->Spawn(tasks...);
	numSpawnedChildren_ = 0;
	numSyncedChildren_ = 0;
      }
      inline void Spawn(Task* task){
	task->SetParent(this);
	numSpawnedChildren_ = 1;
	schedulerHandle_->Spawn(task);
	numSpawnedChildren_ = 0;
	numSyncedChildren_ = 0;
      }
      template<class TaskView>
      inline void Spawn(TaskView& taskView){
	SplitTask<TaskView> st(&taskView);
	Task* t = static_cast<Task*>(&st);
	Spawn(t);
      }
      //helper function to allow setting the parent of each Task in a type pack 
      //recursion basis
      template<class... Tasks>
      inline void SetParentsToThis(){}

      template<class... Tasks>
      inline void SetParentsToThis(Task* task,Tasks... tasks){
	if (task != 0)
	  task->SetParent(this);
	SetParentsToThis(tasks...);
      }
      template<class... Tasks>
      inline void SetParentsToThis(Task& task,Tasks... tasks){
	task.SetParent(this);
	SetParentsToThis(tasks...);
      }
      inline Task* GetParent() {
	return parent_;
      }
      inline bool IsReady() {
	return numSyncedChildren_ == numSpawnedChildren_;
      }
      inline void SetSchedulerHandle(Handle* handle){
	schedulerHandle_ = handle;
      }
      inline void SetParent(Task* p){
	parent_ = p;
      }
    protected:
      inline void IncNumSyncedChildren(){
	++numSyncedChildren_;
      }
      tbb::atomic<unsigned int> numSyncedChildren_;
      unsigned int numSpawnedChildren_;
      Task* parent_;
      Handle* schedulerHandle_;
    };

    template<class View>
    class SplitTask : public Task {
    public:
      SplitTask(View* view) : Task(), view_(view,0,view->size()) {
      }
      SplitTask(View& view) : Task(), view_(&view,0,view.size()) {
      }
      SplitTask(SubView<View>& view) : Task(), view_(view) {
      }
      inline void operator()(){
	if (view_.size() > 3) { 
	  typename View::size_type middle = view_.size()/2;
	  SubView<View> firstInputHalf(view_,0,middle);
	  SubView<View> secondInputHalf(view_,middle,view_.size());
	  SplitTask<View> first(firstInputHalf);
	  SplitTask<View> second(secondInputHalf);
	  Spawn(&first,&second);
	} else if(view_.size() == 3) {
	  Spawn(view_[0],view_[1],view_[2]);
	} else if(view_.size() == 2){
	  Spawn(view_[0],view_[1]);
	} else if(view_.size() == 1){
	  Spawn(dynamic_cast<Task*>(view_[0]));
	} else {
	  //It make no sense to reach this;
	  throw 9;
	}
      }
      friend std::ostream& operator<<(std::ostream& out, const SplitTask<View>& st){
	st.Print(out);
	return out;
      }
      void Print(std::ostream& out) const {
	out<<"SplitView("<<view_<<")";
      }
    private:
      SubView<View> view_;
    };
  };

} //end namespace mcl

#endif //MCL_WS_SCHEDULER
